// Copyright 2018 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package sw64

import (
	"cmd/internal/obj"
	"cmd/internal/objabi"
	"math"
)

type builderFn func(c ctxt77, p *obj.Prog) []uint32

type Optab struct {
	as   obj.As
	a1   uint8
	a2   uint8
	a3   uint8
	size int //generated instructions length in byte unit

	builder builderFn
}

const (
	C_NONE  = iota
	C_RREG  // integer register
	C_FREG  // float-point register
	C_REG   // any register
	C_MCON  // integer constant [-0x8000,0x7fff]
	C_SUCON // integer constant not greater than 0xff
	C_LCON  // integer constant without restriction
	C_FCON  // float-point constant without restriction
	C_LEXT  // external symbol
	C_ADDR  // all symbol
	C_SAUTO // (a.NAME == NAME_PARAM || a.NAME == NAME_AUTO) && offset is [-0x7fff, 0x7fff]
	C_LAUTO // a.NAME == NAME_PARAM || a.NAME == NAME_AUTO
	C_S0REG // (a.NAME == NAME_NONE) && offset is [-0x7fff, 0x7fff]
	C_L0REG // a.NAME == NAME_NONE
	C_TLS

	C_NCLASS /* must be the last */
)

func matchClass(a obj.Addr, c uint8) bool {
	switch c {
	case C_TLS:
		return a.Sym != nil && a.Sym.Type == objabi.STLSBSS
	case C_NONE:
		return a.Type == obj.TYPE_NONE
	case C_REG:
		return a.Type == obj.TYPE_REG && IsReg(int(a.Reg))
	case C_RREG:
		return a.Type == obj.TYPE_REG && IsRReg(int(a.Reg))
	case C_FREG:
		return a.Type == obj.TYPE_REG && IsFReg(int(a.Reg))
	case C_MCON:
		return a.Type == obj.TYPE_CONST && isint16(a.Offset)
	case C_SUCON:
		return a.Type == obj.TYPE_CONST && isuint8(uint64(a.Offset))
	case C_LCON:
		return a.Type == obj.TYPE_CONST
	case C_FCON:
		return a.Type == obj.TYPE_FCONST
	case C_LEXT:
		switch a.Type {
		case obj.TYPE_ADDR, obj.TYPE_BRANCH,
			obj.TYPE_INDIR, obj.TYPE_MEM:
			switch a.Name {
			case obj.NAME_EXTERN, obj.NAME_STATIC:
				return true
			default:
				return false
			}
		default:
			return false
		}
	case C_ADDR:
		return a.Type == obj.TYPE_ADDR || a.Type == obj.TYPE_MEM
	case C_SAUTO:
		return isauto(a) && isint16(a.Offset)
	case C_LAUTO:
		return isauto(a) && (a.Offset > math.MaxInt16 ||
			-a.Offset > math.MaxInt16)
	case C_S0REG:
		return isgenericReg(a) && isint16(a.Offset)
	case C_L0REG:
		return isgenericReg(a) && (a.Offset > math.MaxInt16 ||
			-a.Offset > math.MaxInt16)
	default:
		panic("unknown class type")
	}
}

func isgenericReg(a obj.Addr) bool {
	return (a.Type == obj.TYPE_ADDR || a.Type == obj.TYPE_MEM) &&
		(a.Name == obj.NAME_NONE)
}

func isauto(a obj.Addr) bool {
	return (a.Type == obj.TYPE_ADDR || a.Type == obj.TYPE_MEM) &&
		(a.Name == obj.NAME_PARAM || a.Name == obj.NAME_AUTO)
}

const noOpIndex uint16 = 1<<16 - 1

var optab = []Optab{
	/* OPCODE, from, prog->RestArgs[0], to, size */
	{ALDI, C_RREG, C_NONE, C_MCON, 4, buildInsn},

	{ASTB, C_RREG, C_NONE, C_TLS, 12, buildTLS},
	{ASTH, C_RREG, C_NONE, C_TLS, 12, buildTLS},
	{ASTW, C_RREG, C_NONE, C_TLS, 12, buildTLS},
	{ASTL, C_RREG, C_NONE, C_TLS, 12, buildTLS},
	{ALDBU, C_RREG, C_NONE, C_TLS, 12, buildTLS},
	{ALDHU, C_RREG, C_NONE, C_TLS, 12, buildTLS},
	{ALDW, C_RREG, C_NONE, C_TLS, 12, buildTLS},
	{ALDL, C_RREG, C_NONE, C_TLS, 12, buildTLS},

	{ALDI, C_RREG, C_NONE, C_SAUTO, 4, buildAuto},
	{ALDI, C_RREG, C_NONE, C_LAUTO, 20, buildAuto},
	{ALDBU, C_RREG, C_NONE, C_SAUTO, 4, buildAuto},
	{ALDBU, C_RREG, C_NONE, C_LAUTO, 20, buildAuto},
	{ALDHU, C_RREG, C_NONE, C_SAUTO, 4, buildAuto},
	{ALDHU, C_RREG, C_NONE, C_LAUTO, 20, buildAuto},
	{ALDW, C_RREG, C_NONE, C_SAUTO, 4, buildAuto},
	{ALDW, C_RREG, C_NONE, C_LAUTO, 20, buildAuto},
	{ALDL, C_RREG, C_NONE, C_SAUTO, 4, buildAuto},
	{ALDL, C_RREG, C_NONE, C_LAUTO, 20, buildAuto},

	{ASTB, C_RREG, C_NONE, C_SAUTO, 4, buildAuto},
	{ASTB, C_RREG, C_NONE, C_LAUTO, 20, buildAuto},
	{ASTH, C_RREG, C_NONE, C_SAUTO, 4, buildAuto},
	{ASTH, C_RREG, C_NONE, C_LAUTO, 20, buildAuto},
	{ASTW, C_RREG, C_NONE, C_SAUTO, 4, buildAuto},
	{ASTW, C_RREG, C_NONE, C_LAUTO, 20, buildAuto},
	{ASTL, C_RREG, C_NONE, C_SAUTO, 4, buildAuto},
	{ASTL, C_RREG, C_NONE, C_LAUTO, 20, buildAuto},

	{AFLDD, C_FREG, C_NONE, C_SAUTO, 4, buildAuto},
	{AFLDS, C_FREG, C_NONE, C_SAUTO, 4, buildAuto},
	{AFSTS, C_FREG, C_NONE, C_SAUTO, 4, buildAuto},
	{AFSTD, C_FREG, C_NONE, C_SAUTO, 4, buildAuto},
	{AFLDD, C_FREG, C_NONE, C_LAUTO, 20, buildAuto},
	{AFLDS, C_FREG, C_NONE, C_LAUTO, 20, buildAuto},
	{AFSTS, C_FREG, C_NONE, C_LAUTO, 20, buildAuto},
	{AFSTD, C_FREG, C_NONE, C_LAUTO, 20, buildAuto},

	{ALDI, C_RREG, C_NONE, C_S0REG, 4, buildAuto},
	{ALDI, C_RREG, C_NONE, C_L0REG, 20, buildAuto},
	{ALDBU, C_RREG, C_NONE, C_S0REG, 4, buildAuto},
	{ALDBU, C_RREG, C_NONE, C_L0REG, 20, buildAuto},
	{ALDHU, C_RREG, C_NONE, C_S0REG, 4, buildAuto},
	{ALDHU, C_RREG, C_NONE, C_L0REG, 20, buildAuto},
	{ALDW, C_RREG, C_NONE, C_S0REG, 4, buildAuto},
	{ALDW, C_RREG, C_NONE, C_L0REG, 20, buildAuto},
	{ALDL, C_RREG, C_NONE, C_S0REG, 4, buildAuto},
	{ALDL, C_RREG, C_NONE, C_L0REG, 20, buildAuto},

	{ASTB, C_RREG, C_NONE, C_S0REG, 4, buildAuto},
	{ASTB, C_RREG, C_NONE, C_L0REG, 20, buildAuto},
	{ASTH, C_RREG, C_NONE, C_S0REG, 4, buildAuto},
	{ASTH, C_RREG, C_NONE, C_L0REG, 20, buildAuto},
	{ASTW, C_RREG, C_NONE, C_S0REG, 4, buildAuto},
	{ASTW, C_RREG, C_NONE, C_L0REG, 20, buildAuto},
	{ASTL, C_RREG, C_NONE, C_S0REG, 4, buildAuto},
	{ASTL, C_RREG, C_NONE, C_L0REG, 20, buildAuto},

	{AFLDD, C_FREG, C_NONE, C_S0REG, 4, buildAuto},
	{AFLDS, C_FREG, C_NONE, C_S0REG, 4, buildAuto},
	{AFSTS, C_FREG, C_NONE, C_S0REG, 4, buildAuto},
	{AFSTD, C_FREG, C_NONE, C_S0REG, 4, buildAuto},
	{AFLDD, C_FREG, C_NONE, C_L0REG, 20, buildAuto},
	{AFLDS, C_FREG, C_NONE, C_L0REG, 20, buildAuto},
	{AFSTS, C_FREG, C_NONE, C_L0REG, 20, buildAuto},
	{AFSTD, C_FREG, C_NONE, C_L0REG, 20, buildAuto},

	{obj.AJMP, C_RREG, C_NONE, C_LEXT, 12, buildCall},
	{obj.ACALL, C_RREG, C_NONE, C_LEXT, 12, buildCall},
	{ABR, C_RREG, C_NONE, C_LEXT, 4, buildCall},
	{ABR, C_NONE, C_NONE, C_LEXT, 4, buildCall},

	{obj.AJMP, C_RREG, C_NONE, C_ADDR, 4, buildCallByReg},
	{obj.ACALL, C_RREG, C_NONE, C_ADDR, 4, buildCallByReg},

	{ASUBL, C_RREG, C_SUCON, C_RREG, 4, buildInsn}, //must not be deleted
	{AADDL, C_RREG, C_SUCON, C_RREG, 4, buildInsn},
	{AADDW, C_RREG, C_SUCON, C_RREG, 4, buildInsn},
	{AMULW, C_RREG, C_SUCON, C_RREG, 4, buildInsn},
	{AMULL, C_RREG, C_SUCON, C_RREG, 4, buildInsn},
	{AUMULH, C_RREG, C_SUCON, C_RREG, 4, buildInsn},
	{ASUBL, C_RREG, C_MCON, C_RREG, 8, buildLargeImmediate},
	{AADDL, C_RREG, C_MCON, C_RREG, 8, buildLargeImmediate},
	{AADDW, C_RREG, C_MCON, C_RREG, 8, buildLargeImmediate},
	{AMULW, C_RREG, C_MCON, C_RREG, 8, buildLargeImmediate},
	{AMULL, C_RREG, C_MCON, C_RREG, 8, buildLargeImmediate},
	{AUMULH, C_RREG, C_MCON, C_RREG, 8, buildLargeImmediate},
	{ASUBL, C_RREG, C_LCON, C_RREG, 16, buildLargeImmediate},
	{AADDL, C_RREG, C_LCON, C_RREG, 16, buildLargeImmediate},
	{AADDW, C_RREG, C_LCON, C_RREG, 16, buildLargeImmediate},
	{AMULW, C_RREG, C_LCON, C_RREG, 16, buildLargeImmediate},
	{AMULL, C_RREG, C_LCON, C_RREG, 16, buildLargeImmediate},
	{AUMULH, C_RREG, C_LCON, C_RREG, 16, buildLargeImmediate},

	{AAND, C_RREG, C_SUCON, C_RREG, 4, buildInsn}, //must not be deleted
	{AAND, C_RREG, C_MCON, C_RREG, 8, buildLargeImmediate},
	{AAND, C_RREG, C_LCON, C_RREG, 16, buildLargeImmediate},

	{ABIS, C_RREG, C_SUCON, C_RREG, 4, buildInsn}, //must not be deleted
	{ABIS, C_RREG, C_MCON, C_RREG, 8, buildLargeImmediate},
	{ABIS, C_RREG, C_LCON, C_RREG, 16, buildLargeImmediate},

	{AXOR, C_RREG, C_SUCON, C_RREG, 4, buildInsn}, //must not be deleted
	{AXOR, C_RREG, C_MCON, C_RREG, 8, buildLargeImmediate},
	{AXOR, C_RREG, C_LCON, C_RREG, 16, buildLargeImmediate},

	{AORNOT, C_RREG, C_SUCON, C_RREG, 4, buildInsn}, //must not be deleted
	{AORNOT, C_RREG, C_MCON, C_RREG, 8, buildLargeImmediate},
	{AORNOT, C_RREG, C_LCON, C_RREG, 16, buildLargeImmediate},

	{ASLL, C_RREG, C_SUCON, C_RREG, 4, buildInsn}, //must not be deleted
	{ASLL, C_RREG, C_MCON, C_RREG, 8, buildLargeImmediate},
	{ASLL, C_RREG, C_LCON, C_RREG, 16, buildLargeImmediate},

	{ASRL, C_RREG, C_SUCON, C_RREG, 4, buildInsn}, //must not be deleted
	{ASRL, C_RREG, C_MCON, C_RREG, 8, buildLargeImmediate},
	{ASRL, C_RREG, C_LCON, C_RREG, 16, buildLargeImmediate},

	{ASRA, C_RREG, C_SUCON, C_RREG, 4, buildInsn}, //must not be deleted
	{ASRA, C_RREG, C_MCON, C_RREG, 8, buildLargeImmediate},
	{ASRA, C_RREG, C_LCON, C_RREG, 16, buildLargeImmediate},

	{ASEXTB, C_RREG, C_SUCON, C_RREG, 4, buildInsn}, //must not be deleted
	{ASEXTB, C_RREG, C_MCON, C_RREG, 8, buildLargeImmediate},
	{ASEXTB, C_RREG, C_LCON, C_RREG, 16, buildLargeImmediate},

	{ASEXTH, C_RREG, C_SUCON, C_RREG, 4, buildInsn}, //must not be deleted
	{ASEXTH, C_RREG, C_MCON, C_RREG, 8, buildLargeImmediate},
	{ASEXTH, C_RREG, C_LCON, C_RREG, 16, buildLargeImmediate},

	{ACMPEQ, C_RREG, C_SUCON, C_RREG, 4, buildInsn}, //must not be deleted
	{ACMPEQ, C_RREG, C_MCON, C_RREG, 8, buildLargeImmediate},
	{ACMPEQ, C_RREG, C_LCON, C_RREG, 16, buildLargeImmediate},

	{ACMPLT, C_RREG, C_SUCON, C_RREG, 4, buildInsn}, //must not be deleted
	{ACMPLT, C_RREG, C_MCON, C_RREG, 8, buildLargeImmediate},
	{ACMPLT, C_RREG, C_LCON, C_RREG, 16, buildLargeImmediate},

	{ACMPLE, C_RREG, C_SUCON, C_RREG, 4, buildInsn}, //must not be deleted
	{ACMPLE, C_RREG, C_MCON, C_RREG, 8, buildLargeImmediate},
	{ACMPLE, C_RREG, C_LCON, C_RREG, 16, buildLargeImmediate},

	{ACMPULT, C_RREG, C_SUCON, C_RREG, 4, buildInsn}, //must not be deleted
	{ACMPULT, C_RREG, C_MCON, C_RREG, 8, buildLargeImmediate},
	{ACMPULT, C_RREG, C_LCON, C_RREG, 16, buildLargeImmediate},

	{ACMPULE, C_RREG, C_SUCON, C_RREG, 4, buildInsn}, //must not be deleted
	{ACMPULE, C_RREG, C_MCON, C_RREG, 8, buildLargeImmediate},
	{ACMPULE, C_RREG, C_LCON, C_RREG, 16, buildLargeImmediate},

	{ALDBU, C_RREG, C_NONE, C_LEXT, 8, buildLoad},
	{ALDHU, C_RREG, C_NONE, C_LEXT, 8, buildLoad},
	{ALDW, C_RREG, C_NONE, C_LEXT, 8, buildLoad},
	{ALDL, C_RREG, C_NONE, C_LEXT, 8, buildLoad},
	{ASYMADDR, C_RREG, C_NONE, C_LEXT, 8, buildLoad},
	{AFLDD, C_FREG, C_NONE, C_LEXT, 12, buildLoad},
	{AFLDS, C_FREG, C_NONE, C_LEXT, 12, buildLoad},

	{ASTB, C_RREG, C_NONE, C_LEXT, 8, buildStore},
	{ASTH, C_RREG, C_NONE, C_LEXT, 8, buildStore},
	{ASTW, C_RREG, C_NONE, C_LEXT, 8, buildStore},
	{ASTL, C_RREG, C_NONE, C_LEXT, 8, buildStore},
	{AFSTS, C_FREG, C_NONE, C_LEXT, 8, buildStore},
	{AFSTD, C_FREG, C_NONE, C_LEXT, 8, buildStore},

	{AWORD, C_LCON, C_NONE, C_NONE, 4, buildWORD},

	{ALDGP, C_RREG, C_NONE, C_RREG, 16, buildLDGP},
	{ANOOP, C_NONE, C_NONE, C_NONE, 4, buildNoop},

	{obj.AUNDEF, C_NONE, C_NONE, C_NONE, 4, buildTrap},
	{obj.ANOP, C_NONE, C_NONE, C_NONE, 0, buildNull},
	{obj.AFUNCDATA, C_SUCON, C_NONE, C_ADDR, 0, buildNull},
	{obj.APCDATA, C_SUCON, C_NONE, C_LCON, 0, buildNull},
}

// updateOpIndex set the value of p.Optab to index + 1
// the index has two special value
//   1. zero is meanings the p hasn't been initialized, so we need find the value.
//   2. noOpIndex is meanings the p is a normal instructions e.g. no need specially
//      handling and it's size is 4.
func progParseOpIndex(p *obj.Prog) uint16 {
	if p.Optab != 0 {
		return p.Optab
	}

	for i, o := range optab {
		if o.as != p.As {
			continue
		}

		if len(p.RestArgs) != 0 && !matchClass(p.RestArgs[0], o.a2) {
			continue
		} else if len(p.RestArgs) == 0 && o.a2 != C_NONE {
			continue
		}
		if matchClass(p.From, o.a1) &&
			matchClass(p.To, o.a3) {
			p.Optab = uint16(i + 1)
			return p.Optab
		}
	}
	p.Optab = noOpIndex
	if p.To.Sym != nil {
		p.Ctxt.Diag("TODO handle this! %v \n\tFrom:%v\n\tTo:%v", p, p.From, p.To)
	}
	return noOpIndex
}

func progBuilder(p *obj.Prog) builderFn {
	index := p.Optab

	if (p.As == ALSTW || p.As == ALSTL) && p.Pc&0x7 != 0 {
		return buildUnalignedAtomic
	}

	if index == noOpIndex {
		return buildInsn
	}
	return optab[index-1].builder
}

func progSize(p *obj.Prog) int64 {
	index := p.Optab

	if (p.As == ALSTW || p.As == ALSTL) && p.Pc&0x7 != 0 {
		return 8
	}

	if index == noOpIndex {
		return 4
	}
	return int64(optab[index-1].size)
}

func IsReg(r int) bool {
	return REG_R0 <= r && r <= REG_F31
}
func IsFReg(r int) bool {
	return REG_F0 <= r && r <= REG_F31
}
func IsRReg(r int) bool {
	return REG_R0 <= r && r <= REG_R31
}

func isint16(v int64) bool {
	return int64(int16(v)) == v
}
func isuint8(v uint64) bool {
	return uint64(uint8(v)) == v
}
func isint32(v int64) bool {
	return int64(int32(v)) == v
}
func isuint32(v uint64) bool {
	return uint64(uint32(v)) == v
}

func buildop(ctxt *obj.Link) {
	// We find the index by iterating the optab now,
	// so nothing todo now.
	// snyh_TODO: init a oprange for speed up this process
}
