// Copyright 2018 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package sw64

import (
	"cmd/internal/obj"
	"cmd/internal/objabi"
	"cmd/internal/sys"
	"fmt"
)

func init() {
	obj.RegisterRegister(obj.RBaseSW64, REG_R0+1024, rconv)
	obj.RegisterOpcode(obj.ABaseSW64, Anames)
}

// ctxt77 holds state while assembling a single function.
// Each function gets a fresh ctxt77.
// This allows for multiple functions to be safely concurrently assembled.
type ctxt77 struct {
	ctxt       *obj.Link
	newprog    obj.ProgAlloc
	cursym     *obj.LSym
	autosize   int32
	instoffset int64
	pc         int64
}

func rconv(r int) string {
	switch {
	case 0 == r:
		return "NONE"
	case REG_R15 == r:
		return "g"
	case REG_R29 == r:
		return "RSB"
	case REG_R30 == r:
		return "SP"
	case REGCTXT == r:
		return "REGCTXT"
	case REGZERO == r:
		return "ZERO"
	case IsRReg(r):
		return fmt.Sprintf("R%d", r-REG_R0)
	case IsFReg(r):
		return fmt.Sprintf("F%d", r-REG_F0)
	}
	return fmt.Sprintf("badreg(%d)", r)
}

func (c *ctxt77) stacksplit(p *obj.Prog, framesize int32) *obj.Prog {
	if framesize == 0 {
		return p
	}

	p = insertLDGP(p, c.newprog, REGZERO)

	// LDL R1, g_stackguard(g)
	p = obj.Appendp(p, c.newprog)
	begin := p

	p.As = ALDL
	p.From.Type = obj.TYPE_REG
	p.From.Reg = REG_R1

	p.To.Type = obj.TYPE_MEM
	p.To.Reg = REGG
	p.To.Offset = 2 * int64(c.ctxt.Arch.PtrSize) // G.stackguard0
	if c.cursym.CFunc() {
		p.To.Offset = 3 * int64(c.ctxt.Arch.PtrSize) // G.stackguard1
	}

	var q *obj.Prog
	if framesize <= objabi.StackSmall {
		// small stack: SP < stackguard
		// CMPULE SP, stackguard, R1
		p = obj.Appendp(p, c.newprog)

		p.As = ACMPULE
		p.From.Type = obj.TYPE_REG
		p.From.Reg = REGSP
		p.SetFrom3(obj.Addr{
			Type: obj.TYPE_REG,
			Reg:  REG_R1,
		})
		p.To.Type = obj.TYPE_REG
		p.To.Reg = REG_R1

	} else if framesize <= objabi.StackBig {
		// large stack: SP-framesize < stackguard-StackSmall
		// SUBL SP, $(framesize-StackSmall), R2
		// CMPULE R2, stackgurad, R1
		p = obj.Appendp(p, c.newprog)

		p.As = ASUBL
		p.From.Type = obj.TYPE_REG
		p.From.Reg = REGSP

		p.SetFrom3(obj.Addr{
			Type:   obj.TYPE_CONST,
			Offset: (int64(framesize) - objabi.StackSmall),
		})
		p.To.Type = obj.TYPE_REG
		p.To.Reg = REG_R2

		p = obj.Appendp(p, c.newprog)
		p.As = ACMPULE
		p.From.Type = obj.TYPE_REG
		p.From.Reg = REG_R2
		p.SetFrom3(obj.Addr{
			Type: obj.TYPE_REG,
			Reg:  REG_R1,
		})
		p.To.Type = obj.TYPE_REG
		p.To.Reg = REG_R1
	} else {
		// Such a large stack we need to protect against wraparound.
		// If SP is close to zero:
		//	SP-stackguard+StackGuard <= framesize + (StackGuard-StackSmall)
		// The +StackGuard on both sides is required to keep the left side positive:
		// SP is allowed to be slightly below stackguard. See stack.h.
		//
		// Preemption sets stackguard to StackPreempt, a very large value.
		// That breaks the math above, so we have to check for that explicitly.
		//	// stackguard is R1

		//	LDI R2, $StackPreempt
		p = obj.Appendp(p, c.newprog)
		p.As = ALDI
		p.To.Type = obj.TYPE_CONST
		p.To.Offset = objabi.StackPreempt
		p.From.Type = obj.TYPE_REG
		p.From.Reg = REG_R2

		// CMPEQ R1, R2, R0
		// BNE R0, label-of-call-to-morestack
		p = obj.Appendp(p, c.newprog)
		p.As = ACMPEQ
		p.From.Type, p.From.Reg = obj.TYPE_REG, REG_R1
		p.SetFrom3(obj.Addr{Type: obj.TYPE_REG, Reg: REG_R2})
		p.To.Type, p.To.Reg = obj.TYPE_REG, REG_R0
		p = obj.Appendp(p, c.newprog)
		q = p
		p.As = ABNE
		p.From.Type, p.From.Reg = obj.TYPE_REG, REG_R0
		p.To.Type = obj.TYPE_BRANCH

		//	ADDL SP, $StackGuard, R2
		p = obj.Appendp(p, c.newprog)
		p.As = AADDL
		p.From.Type, p.From.Reg = obj.TYPE_REG, REGSP
		p.SetFrom3(obj.Addr{Type: obj.TYPE_CONST, Offset: objabi.StackGuard})
		p.To.Type, p.To.Reg = obj.TYPE_REG, REG_R2

		//	SUBL R2, R1, R2
		p = obj.Appendp(p, c.newprog)
		p.As = ASUBL
		p.From.Type, p.From.Reg = obj.TYPE_REG, REG_R2
		p.SetFrom3(obj.Addr{Type: obj.TYPE_REG, Reg: REG_R1})
		p.To.Type, p.To.Reg = obj.TYPE_REG, REG_R2

		//	LDI R1, $(framesize+(StackGuard-StackSmall))
		v := int64(framesize) + objabi.StackGuard - objabi.StackSmall
		hi := int16(v >> 16)
		lo := int16(v & 0xffff)
		if lo < 0 {
			hi = hi + 1
			lo = int16(v - int64(hi)<<16)
		}
		p = obj.Appendp(p, c.newprog)
		p.As = ALDIH
		p.From.Type, p.From.Reg = obj.TYPE_REG, REG_R1
		p.To.Type, p.To.Offset = obj.TYPE_CONST, int64(hi)
		p = obj.Appendp(p, c.newprog)
		p.As = ALDI
		p.From.Type, p.From.Reg = obj.TYPE_REG, REG_R1
		p.To.Type, p.To.Offset, p.To.Reg = obj.TYPE_ADDR, int64(lo), REG_R1

		// CMPULE R2, R1, R1
		p = obj.Appendp(p, c.newprog)
		p.As = ACMPULE
		p.From.Type, p.From.Reg = obj.TYPE_REG, REG_R2
		p.SetFrom3(obj.Addr{Type: obj.TYPE_REG, Reg: REG_R1})
		p.To.Type, p.To.Reg = obj.TYPE_REG, REG_R1
	}

	// q1: BEQ R1, done
	p = obj.Appendp(p, c.newprog)
	q1 := p

	p.As = ABEQ
	p.From.Type = obj.TYPE_REG
	p.From.Reg = REG_R1
	p.To.Type = obj.TYPE_BRANCH

	// LDI R3, LINK
	p = obj.Appendp(p, c.newprog)
	p.As = ALDI
	p.From.Type = obj.TYPE_REG
	p.From.Reg = REG_R3
	p.To.Type = obj.TYPE_REG
	p.To.Reg = REGLINK
	if q != nil {
		q.To.Val = p
	}

	// JMP runtime.morestack(SB)
	p = obj.Appendp(p, c.newprog)

	p.As = AJMP
	p.From.Type = obj.TYPE_REG
	//Don't use RA, otherwise the GP value will wrong because when
	//the morestack return the RA is not the next instruction address.
	p.From.Reg = REG_R2
	p.To.Type = obj.TYPE_BRANCH
	p.To.Name = obj.NAME_EXTERN

	if c.cursym.CFunc() {
		p.To.Sym = c.ctxt.Lookup("runtime.morestackc")
	} else if !c.cursym.Func.Text.From.Sym.NeedCtxt() {
		p.To.Sym = c.ctxt.Lookup("runtime.morestack_noctxt")
	} else {
		p.To.Sym = c.ctxt.Lookup("runtime.morestack")
	}

	// JMP	start
	p = obj.Appendp(p, c.newprog)
	p.As = ABR
	p.From.Type, p.From.Reg = obj.TYPE_REG, REGZERO
	p.To.Type, p.To.Val = obj.TYPE_BRANCH, begin

	// placeholder for q1's jump target
	p = obj.Appendp(p, c.newprog)
	p.As = obj.ANOP // zero-width place holder
	q1.To.Val = p

	return p
}

func insertLDGP(q *obj.Prog, newprog obj.ProgAlloc, rip int16) *obj.Prog {
	q = obj.Appendp(q, newprog)
	q.As = ALDGP
	q.From = obj.Addr{Type: obj.TYPE_REG, Reg: REG_R29}
	q.To = obj.Addr{Type: obj.TYPE_REG, Reg: rip}
	return q
}

func preprocess(ctxt *obj.Link, cursym *obj.LSym, newprog obj.ProgAlloc) {
	// 1. handle ARET and ATEXT to setup function prologue and epilogue
	// 2. adjust Prog.Spadj if the prog change hardward SP

	c := ctxt77{ctxt: ctxt, newprog: newprog, cursym: cursym}
	p := c.cursym.Func.Text
	textstksiz := p.To.Offset

	c.cursym.Func.Args = p.To.Val.(int32)
	c.cursym.Func.Locals = int32(textstksiz)

	if textstksiz < 0 {
		ctxt.Diag("Use NOFRAME attribute instead of a negative frame size.\n%v\n", p)
		return
	}
	autosize := int32(textstksiz) + int32(ctxt.FixedFrameSize())
	if cursym.Attribute.NoFrame() {
		if textstksiz != 0 {
			ctxt.Diag("NoFrame symbol %v with non zero frame size(%d)", cursym, textstksiz)
		}
		autosize = 0
	}

	//convert virtual SP to hardware SP
	//see also: https://bj.git.sndu.cn/xiabin/go-sw64/wikis/stack-frame-layout
	for p := cursym.Func.Text; p != nil; p = p.Link {
		switch p.To.Name {
		case obj.NAME_PARAM:
			p.To.Type = obj.TYPE_ADDR
			p.To.Reg = REGSP
			p.To.Offset += int64(autosize) + c.ctxt.FixedFrameSize()
			p.To.Sym = nil
		case obj.NAME_AUTO:
			p.To.Type = obj.TYPE_ADDR
			p.To.Reg = REGSP
			p.To.Offset += int64(autosize)
			p.To.Sym = nil
		}
	}

	for p := cursym.Func.Text; p != nil; p = p.Link {
		o := p.As
		switch o {
		case obj.ATEXT:
			q := p
			if !p.From.Sym.NoSplit() {
				q = c.stacksplit(q, autosize) // emit split check
			}
			if autosize != 0 {
				// LDI SP, $-autosize(SP)
				q = obj.Appendp(q, newprog)
				q.As = ALDI
				q.From.Type = obj.TYPE_REG
				q.From.Reg = REGSP
				q.To = obj.Addr{
					Type:   obj.TYPE_ADDR,
					Reg:    REGSP,
					Offset: int64(-autosize),
				}

				// STL RA, $0(SP)
				q = obj.Appendp(q, newprog)
				q.Spadj = autosize
				q.As = ASTL
				q.From.Type = obj.TYPE_REG
				q.From.Reg = REGLINK
				q.To.Type = obj.TYPE_MEM
				q.To.Reg = REGSP
			}

			//_snyh_TODO: see #131
			q = insertLDGP(q, newprog, REGZERO)

			if c.cursym.Func.Text.From.Sym.Wrapper() {
				// if(g->panic != nil && g->panic->argp == FP)
				//    g->panic->argp = bottom-of-frame

				// LDL R1, g_panic(g)
				q = obj.Appendp(q, newprog)
				q.As = ALDL
				q.From = obj.Addr{Type: obj.TYPE_REG, Reg: REG_R1}
				q.To = obj.Addr{
					Type:   obj.TYPE_MEM,
					Reg:    REGG,
					Offset: 4 * int64(c.ctxt.Arch.PtrSize), // G.panic
				}

				// BEQ R1, end
				q = obj.Appendp(q, newprog)
				q.As = ABEQ
				q.From.Type = obj.TYPE_REG
				q.From.Reg = REG_R1
				q.To.Type = obj.TYPE_BRANCH
				p1 := q

				// LDL R2, panic_argp(R1)
				q = obj.Appendp(q, newprog)
				q.As = ALDL
				q.From = obj.Addr{Type: obj.TYPE_REG, Reg: REG_R2}
				q.To = obj.Addr{
					Type:   obj.TYPE_MEM,
					Reg:    REG_R1,
					Offset: 0, // Panic.argp
				}

				// ADDL SP, $autosize+FixedFrameSize, R3
				q = obj.Appendp(q, newprog)
				q.As = AADDL
				q.From.Type = obj.TYPE_REG
				q.From.Reg = REGSP
				q.SetFrom3(obj.Addr{
					Type:   obj.TYPE_CONST,
					Offset: int64(autosize) + ctxt.FixedFrameSize(),
				})
				q.To.Type = obj.TYPE_REG
				q.To.Reg = REG_R3

				// CMPEQ R2, R3, R0
				q = obj.Appendp(q, newprog)
				q.As = ACMPEQ
				q.From.Type = obj.TYPE_REG
				q.From.Reg = REG_R2
				q.SetFrom3(obj.Addr{
					Type: obj.TYPE_REG,
					Reg:  REG_R3,
				})
				q.To.Type = obj.TYPE_REG
				q.To.Reg = REG_R0
				// BEQ R0, end
				q = obj.Appendp(q, newprog)
				q.As = ABEQ
				q.From.Type = obj.TYPE_REG
				q.From.Reg = REG_R0
				q.To.Type = obj.TYPE_BRANCH
				p2 := q

				// ADD SP, $FIXED_FRAME, R2
				q = obj.Appendp(q, newprog)
				q.As = AADDL
				q.From.Type = obj.TYPE_REG
				q.From.Reg = REGSP
				q.SetFrom3(obj.Addr{
					Type:   obj.TYPE_CONST,
					Offset: ctxt.FixedFrameSize(),
				})
				q.To.Type = obj.TYPE_REG
				q.To.Reg = REG_R2

				// STL R2, panic_argp(R1)
				q = obj.Appendp(q, newprog)
				q.As = ASTL
				q.From.Type = obj.TYPE_REG
				q.From.Reg = REG_R2
				q.To.Type = obj.TYPE_MEM
				q.To.Reg = REG_R1
				q.To.Offset = 0 // Panic.argp

				// end:
				q = obj.Appendp(q, newprog)
				q.As = obj.ANOP
				p1.To.Val = q
				p2.To.Val = q
			}
			p = q

		case obj.ARET:
			if autosize != 0 {
				tmp := newprog()
				*tmp = *p

				// We can't use q = obj.Appendp(q, newprog),
				// because the RET may be a jump target, and Appendp
				// will change the pointer value of q
				// LDL RA, $0(SP)
				q := p
				q.As = ALDL
				q.From.Type = obj.TYPE_REG
				q.From.Reg = REGLINK
				q.To.Type = obj.TYPE_MEM
				q.To.Reg = REGSP

				//LDI SP, $autosize(SP)
				q = obj.Appendp(q, newprog)
				q.As = ALDI
				q.From.Type = obj.TYPE_REG
				q.From.Reg = REGSP
				q.To = obj.Addr{
					Type:   obj.TYPE_ADDR,
					Reg:    REGSP,
					Offset: int64(autosize),
				}
				q.Spadj = -autosize

				q = obj.Appendp(q, newprog)
				*q = *tmp

				//insert a noop for restore spadj
				q = obj.Appendp(q, newprog)
				q.As = obj.ANOP
				q.Spadj = autosize
				p = q
			}
		}
	}

	for p := cursym.Func.Text; p != nil; p = p.Link {
		switch p.As {
		case obj.ACALL, obj.AJMP:
			p = insertLDGP(p, newprog, REGZERO)
		}
	}
}

func span77(ctxt *obj.Link, cursym *obj.LSym, newprog obj.ProgAlloc) {
	p := cursym.Func.Text
	if p == nil || p.Link == nil { // handle external functions and ELF section symbols
		return
	}
	c := ctxt77{ctxt: ctxt, newprog: newprog, cursym: cursym, autosize: int32(p.To.Offset + ctxt.FixedFrameSize())}

	pc := int64(0)
	p.Pc = pc
	for p = p.Link; p != nil; p = p.Link {
		progParseOpIndex(p)
		p.Pc = pc
		pc += progSize(p)
	}
	cursym.Size = pc
	cursym.Grow(cursym.Size)
	bp := cursym.P

	for p := cursym.Func.Text.Link; p != nil; p = p.Link {
		instructions := progBuilder(p)(c, p)
		if int(progSize(p)/4) != len(instructions) {
			ctxt.Diag("oops bug in builder.\n\t%v\n\tProgSize report %d, but builder output %d\n",
				p, progSize(p)/4, len(instructions))
		}

		for _, mc := range instructions {
			if len(bp) < 4 {
				ctxt.Diag("oops %v\n\t%v %v\n", p, p.To, p.Optab)
				return
			}
			ctxt.Arch.ByteOrder.PutUint32(bp, mc)
			bp = bp[4:]
		}
	}
	verifyLock(c)
}

func verifyLock(c ctxt77) {
	for p := c.cursym.Func.Text.Link; p != nil; p = p.Link {
		// 判断锁装入指令与写锁标志指令是否成对出现
		if p.As == ALLDW || p.As == ALLDL {
			for q := p; q != nil; q = q.Link {
				if q.As != AWR_F {
					if q.As == ALSTW || q.As == ALSTL {
						c.ctxt.Diag("missing WR_F after %s\n", p.As)
						return
					} else {
						if (q.Pc + 4) == c.cursym.Size {
							c.ctxt.Diag("Atomic operation is not logical!\n")
							return
						}
						continue
					}
				} else {
					break
				}
			}
		}

		if p.As == ALSTW || p.As == ALSTL {
			// 判断锁存储指令与读锁标志指令是否成对出现
			if p.Link.As != ARD_F {
				c.ctxt.Diag("missing RD_F after %s\n", p.As)
				return
			}
		}
	}

}

func progedit(ctxt *obj.Link, p *obj.Prog, newprog obj.ProgAlloc) {
	p.From.Class = 0
	p.To.Class = 0
	switch p.As {
	case ACALL:
		if p.From.Type == obj.TYPE_NONE {
			p.From.Type = obj.TYPE_REG
			p.From.Reg = REGLINK
		}
	case AJMP:
		if p.To.Type == obj.TYPE_BRANCH {
			p.As = ABR
		}
		if p.From.Type == obj.TYPE_NONE {
			p.From.Type = obj.TYPE_REG
			p.From.Reg = REGZERO
		}
	case ARET:
		if p.To.Type == obj.TYPE_NONE && p.From.Type == obj.TYPE_NONE {
			p.From.Type = obj.TYPE_REG
			p.From.Reg = REGZERO
			p.To.Type = obj.TYPE_MEM
			p.To.Reg = REGLINK
		}
	case ALDF:
		if p.To.Type == obj.TYPE_FCONST {
			f32 := float32(p.To.Val.(float64))
			p.As = AFLDS
			p.To.Type = obj.TYPE_MEM
			p.To.Sym = ctxt.Float32Sym(f32)
			p.To.Name = obj.NAME_EXTERN
			p.To.Offset = 0
		}
	case ALDI:
		if p.To.Type == obj.TYPE_FCONST {
			f64 := float64(p.To.Val.(float64))
			p.As = AFLDD
			p.To.Type = obj.TYPE_MEM
			p.To.Sym = ctxt.Float64Sym(f64)
			p.To.Name = obj.NAME_EXTERN
			p.To.Offset = 0
		} else if p.To.Type == obj.TYPE_CONST && !isint16(p.To.Offset) {
			p.As = ALDL
			p.To.Type = obj.TYPE_MEM
			p.To.Sym = ctxt.Int64Sym(p.To.Offset)
			p.To.Name = obj.NAME_EXTERN
			p.To.Offset = 0
		}
	case AIFMOVS,
		AIFMOVD,
		AFIMOVS,
		AFIMOVD:
		p.SetFrom3(obj.Addr{
			Type: obj.TYPE_REG,
			Reg:  REGZERO,
		})
	case AFCVTSD,
		AFCVTDS,
		AFCVTDL,
		AFCVTLS,
		AFCVTLD,
		AFCVTLW,
		AFCVTWL,
		AFSQRTS,
		AFSQRTD,
		ACTLZ,
		ACTTZ,
		ACTPOP,
		ASEXTB,
		ASEXTH,
		AFCVTDL_Z,
		AFCVTDL_P,
		AFCVTDL_G,
		AFCVTDL_N:
		p.SetFrom3(p.From)
		p.From.Reg = REGZERO
		p.From.Type = obj.TYPE_REG
	}
}

var LinkSW64 = obj.LinkArch{
	Arch:       sys.ArchSW64,
	Init:       buildop,
	Preprocess: preprocess,
	Assemble:   span77,
	UnaryDst:   unaryDst,
	Progedit:   progedit,
}

//单参数指令，且参数为目的数 ==>(Prog.To = a[0])
//不在这个列表的则参数为源数 ==>(Prog.From = a[0])
var unaryDst = map[obj.As]bool{}
