// Copyright 2018 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package sw64

import (
	"cmd/internal/obj"
	"cmd/internal/objabi"
	"fmt"
)

func buildUnalignedAtomic(c ctxt77, p *obj.Prog) []uint32 {
	o1 := buildNoop(c, c.newprog())
	o2 := buildInsn(c, p)
	return append(o1, o2...)
}

func buildNull(ctxt77, *obj.Prog) []uint32 { return nil }
func buildTrap(ctxt77, *obj.Prog) []uint32 { return []uint32{80} }

func buildTLS(c ctxt77, p *obj.Prog) []uint32 {
	if p.To.Sym == nil || p.From.Type != obj.TYPE_REG || !isint16(p.To.Offset) {
		c.ctxt.Diag("%v is not support TLS", p)
	}

	offset := int16(p.To.Offset)
	rn := int16(p.From.Reg & 31)
	r0 := int16(REG_R0 & 31)
	sym := p.To.Sym

	// sys_call 0x9e
	o1 := OP_SYSCALL(op(ASYS_CALL), 0x9e)
	// LDIH R0, sym(R0) !tprelhi
	o2 := OP_MEMORY(op(ALDIH), r0, r0, 0)
	addrel(c.cursym, int32(p.Pc+4), 0, sym, objabi.R_SW64_TPRELHI)
	// STx/LDx Rn, sym(R0) !tprello
	o3 := OP_MEMORY(op(p.As), rn, r0, offset)
	addrel(c.cursym, int32(p.Pc+8), 0, sym, objabi.R_SW64_TPRELLO)

	return []uint32{o1, o2, o3}
}

func buildNoop(c ctxt77, p *obj.Prog) []uint32 {
	p.As = ALDIH
	p.From.Type = obj.TYPE_REG
	p.From.Reg = REGZERO
	p.To.Type = obj.TYPE_REG
	p.To.Reg = REGSP

	return buildInsn(c, p)
}

func buildLDGP(c ctxt77, p *obj.Prog) []uint32 {
	dummysym := c.cursym
	gp := getRegister(p.From.Reg)
	r := getRegister(p.To.Reg)

	if p.To.Reg == REGZERO {
		// If the ATEXT+0 isn't the noop, then
		// runtime.goexit will be invoked at o3 which is ldgp
		// without setup $at. see the logical in
		// runtime/proc.go' func newproc1 about
		// newg.sched.pc = funcPC(goexit) + sys.PCQuantum
		o1 := OP_MEMORY(0x3f, gp, gp, 0) // only for alignment

		//insert a "BR AT, 0" to find the correct PC value
		o2 := OP_CONTROL(op(ABR), REGTMP, 0)
		o3 := OP_MEMORY(0x3f, gp, REGTMP&31, 0)
		addrel(c.cursym, int32(p.Pc+8), 4, dummysym, objabi.R_SW64_GPDISP)
		o4 := OP_MEMORY(0x3e, gp, gp, 0) //ldi  $gp, m($gp)
		return []uint32{o1, o2, o3, o4}
	} else {
		if p.To.Reg == REG_R27 && p.Pc != 0 {
			p.To.Offset = p.Pc
		}
		o1 := OP_MEMORY(0x3f, gp, r, 0) //ldih $gp, n($r)
		addrel(c.cursym, int32(p.Pc), 4, dummysym, objabi.R_SW64_GPDISP)
		o2 := OP_MEMORY(0x3e, gp, gp, 0) //ldi  $gp, m($gp)

		o3 := OP_MEMORY(0x3e, gp, gp, int16(p.To.Offset))
		o4 := OP_MEMORY(0x3f, gp, gp, 0) // only for alignment
		return []uint32{o1, o2, o3, o4}
	}
}

func buildWORD(c ctxt77, p *obj.Prog) []uint32 {
	if p.As != AWORD {
		c.ctxt.Diag("buildWord not support %v", p)
	}
	return []uint32{uint32(p.From.Offset)}
}

func buildStore(c ctxt77, p *obj.Prog) []uint32 {
	// AFSTD, AFSTS, ASTL, ASTB, ASTH, ASTW:
	sym := p.To.Sym
	if sym == nil {
		c.ctxt.Diag("buildStore not support %v", p)
	}

	// STx Rn, sym(SB)
	r := getRegister(p.From.Reg)
	o1 := OP_MEMORY(0x3f, REGTMP&31, REGSB&31, 0) // ldih TMP, sym(gp)
	addrel(c.cursym, int32(p.Pc), p.To.Offset, sym, objabi.R_SW64_GPRELHIGH)

	o2 := OP_MEMORY(op(p.As), r, REGTMP&31, 0) // STx  Rn, sym_lo(TMP)
	addrel(c.cursym, int32(p.Pc+4), p.To.Offset, sym, objabi.R_SW64_GPRELLOW)
	return []uint32{o1, o2}
}

func buildLoad(c ctxt77, p *obj.Prog) []uint32 {
	// LDx Rn, sym(SB)
	sym := p.To.Sym
	if sym == nil {
		c.ctxt.Diag("buildLoad not support %v", p)
	}

	r := getRegister(p.From.Reg)
	switch p.As {
	case AFLDD, AFLDS:
		o1 := OP_MEMORY(0x3f, REGTMP&31, REGSB&31, 0) // LDIH AT, sym(gp)
		addrel(c.cursym, int32(p.Pc), p.To.Offset, sym, objabi.R_SW64_GPRELHIGH)
		o2 := OP_MEMORY(0x3e, REGTMP&31, REGTMP&31, 0) // LDI  AT, sym_lo(AT)
		addrel(c.cursym, int32(p.Pc+4), p.To.Offset, sym, objabi.R_SW64_GPRELLOW)
		o3 := OP_MEMORY(op(p.As), getRegister(p.From.Reg), REGTMP&31, 0) // FLDX, Fn, (AT)
		return []uint32{o1, o2, o3}
	case ALDL, ALDW, ALDHU, ALDBU, ASYMADDR:
		as := p.As
		if as == ASYMADDR {
			as = ALDI
		}
		o1 := OP_MEMORY(0x3f, r, REGSB&31, 0) // ldih r, sym(gp)
		addrel(c.cursym, int32(p.Pc), p.To.Offset, sym, objabi.R_SW64_GPRELHIGH)
		o2 := OP_MEMORY(op(as), r, r, 0) // ldx  r, sym_lo(r)
		addrel(c.cursym, int32(p.Pc+4), p.To.Offset, sym, objabi.R_SW64_GPRELLOW)
		return []uint32{o1, o2}
	default:
		c.ctxt.Diag("buildLoad doesn't support %v", p)
		return nil
	}
}

// buildLargeImmediate handle #b when it isn't in  [0,0xff].
// if #b is in [0x100, 0x7fff] or [-0x8000,0) then use a extra ALDI to load immediate,
// if #b is in [0x8000, 0xffffffffffffffff] or [-0x8000000000000000, -0x8001] then use a extra
//  symbol to handle it.
//1.set all generated prog's Pc to p.Pc;
//2.keep p.Pc
func buildLargeImmediate(c ctxt77, p *obj.Prog) []uint32 {
	ib := p.RestArgs[0].Offset
	var o1 []uint32
	if isint16(ib) {
		//LDI AT, $const
		q := c.newprog()
		q.Pc = p.Pc
		q.As = ALDI
		q.From.Type = obj.TYPE_REG
		q.From.Reg = REGTMP
		q.To.Type = obj.TYPE_CONST
		q.To.Offset = ib
		o1 = buildInsn(c, q)
	} else {
		//SYMADDR AT, $int64.ib
		q := c.newprog()
		q.Pc = p.Pc

		q.As = ASYMADDR
		q.From.Type = obj.TYPE_REG
		q.From.Reg = REGTMP
		q.To.Type = obj.TYPE_MEM
		q.To.Sym = c.ctxt.Int64Sym(ib)
		q.To.Name = obj.NAME_EXTERN
		o1 = buildLoad(c, q)

		// LDL AT, (AT)
		q2 := c.newprog()
		q2.As = ALDL
		q2.Pc = p.Pc + 8
		q2.From.Type = obj.TYPE_REG
		q2.From.Reg = REGTMP
		q2.To.Type = obj.TYPE_MEM
		q2.To.Reg = REGTMP
		o1 = append(o1, buildInsn(c, q2)...)
	}

	// XXX Rn1, AT, Rn2
	p.SetFrom3(obj.Addr{
		Type: obj.TYPE_REG,
		Reg:  REGTMP,
	})
	o2 := buildInsn(c, p)
	return append(o1, o2...)
}

func buildCallByReg(c ctxt77, p *obj.Prog) []uint32 {
	offset := p.To.Offset
	if (p.As != obj.AJMP && p.As != obj.ACALL) || p.To.Sym != nil || !isint16(offset) {
		c.ctxt.Diag("buildCallByReg doesn't support %v", p)
		return nil
	}
	ra := getRegister(p.From.Reg)
	pv := getRegister(p.To.Reg)
	o1 := OP_MEMORY(op(p.As), ra, pv, int16(offset)) //CALL/JMP RA, (Rn)
	// snyh_TODO: support R_CALLIND
	//addrel(c.cursym, int32(p.Pc), 0, dummysym, objabi.R_CALLIND)
	return []uint32{o1}
}

func buildCall(c ctxt77, p *obj.Prog) []uint32 {
	sym := p.To.Sym
	switch p.As {
	case obj.AJMP, obj.ACALL:
		r := getRegister(p.From.Reg)
		pv := REG_R27 & int16(31)
		o1 := OP_MEMORY(0x3f, pv, REGSB&31, 0) // ldih R27, sym(GP)
		addrel(c.cursym, int32(p.Pc), p.To.Offset, sym, objabi.R_SW64_GPRELHIGH)
		o2 := OP_MEMORY(0x3e, pv, pv, 0) // ldi  R27, sym_lo(R27)
		addrel(c.cursym, int32(p.Pc+4), p.To.Offset, sym, objabi.R_SW64_GPRELLOW)

		o3 := OP_MEMORY(op(p.As), r, pv, 0) //CALL/JMP R, (R27)
		addrel(c.cursym, int32(p.Pc+8), p.To.Offset, sym, objabi.R_SW64_HINT)
		return []uint32{o1, o2, o3}
	case ABSR, ABR,
		ABEQ, ABGE, ABGT, ABLE, ABLT,
		ABNE, ABLBC, ABLBS, AFBEQ, AFBGE, AFBGT, AFBLE, AFBLT, AFBNE:
		addrel(c.cursym, int32(p.Pc), p.To.Offset, sym, objabi.R_SW64_BRADDR)

		var o1 uint32
		if p.To.Type == obj.TYPE_BRANCH {
			target := p.To.Val.(*obj.Prog)
			offset := (target.Pc - p.Pc - 4)
			o1 = OP_CONTROL(op(p.As), getRegister(p.From.Reg), int32(offset)/4)
		} else {
			o1 = OP_CONTROL(op(p.As), getRegister(p.From.Reg), int32(p.To.Offset)/4)
		}
		return []uint32{o1}
	default:
		c.ctxt.Diag("buildCall doesn't support %v", p)
		return nil
	}
}

//1.set all generated prog's Pc to p.Pc;
//2.keep p.Pc
func buildAuto(c ctxt77, p *obj.Prog) []uint32 {
	disp := p.To.Offset
	if p.To.Type != obj.TYPE_MEM && p.To.Type != obj.TYPE_ADDR {
		c.ctxt.Diag("buildLoadAuto not support, type mismatched\n %v", p)
	}
	if isint16(disp) {
		return []uint32{
			OP_MEMORY(op(p.As),
				getRegister(p.From.Reg),
				getRegister(p.To.Reg),
				int16(disp)),
		}
	} else {
		// LDx/STx   Ra, off(Rb)
		//  convert to --->
		// ADDL      Rb, $off, REGTMP  #len 16
		// LDx/STx   Ra, (REGTMP)       #len 4
		rb := p.To.Reg
		off := p.To.Offset

		// ADDL Rb, $off, REGTMP
		addl := c.newprog()
		addl.Pc = p.Pc
		addl.As = AADDL
		addl.From = obj.Addr{
			Type: obj.TYPE_REG,
			Reg:  rb,
		}
		addl.SetFrom3(obj.Addr{
			Type:   obj.TYPE_CONST,
			Offset: off,
		})
		addl.To = obj.Addr{
			Type: obj.TYPE_REG,
			Reg:  REGTMP,
		}

		o1 := buildLargeImmediate(c, addl)

		// LDx Ra, (REGTMP)
		p.To = obj.Addr{
			Type: obj.TYPE_MEM,
			Reg:  REGTMP,
		}
		o2 := buildInsn(c, p)
		return append(o1, o2...)
	}
}

func buildInsn(c ctxt77, p *obj.Prog) []uint32 {
	switch {
	case p.To.Sym != nil,
		p.From.Sym != nil,
		p.As == obj.AJMP,
		p.As == obj.ACALL:
		c.ctxt.Diag("Use special builder to handle symbol. %v\n\t%v", p, p.To)
		return nil
	}
	op, err := findOPTable(p)
	if err != nil {
		c.ctxt.Diag("Something wrong: %v.\n\t%v\n", err, p)
		return nil
	}
	o1 := uint32(0)
	switch op.Class {
	case OPC_CONTROL:
		if p.To.Type == obj.TYPE_BRANCH {
			target := p.To.Val.(*obj.Prog)
			offset := (target.Pc - p.Pc - 4)
			o1 = OP_CONTROL(op.OPCode,
				getRegister(p.From.Reg),
				int32(offset)/4)
		} else {
			o1 = OP_CONTROL(op.OPCode,
				getRegister(p.From.Reg),
				int32(p.To.Offset)/4)
		}
	case OPC_SYSCALL:
		if p.As == ASYS_CALL_B {
			o1 = OP_SYSCALL(op.OPCode, uint32(p.From.Offset))
		} else if p.As == ASYS_CALL {
			o1 = OP_SYSCALL(op.OPCode, uint32(p.From.Offset)|1<<25)
		} else {
			p.Ctxt.Diag("sw64 todo implement %v", op)
		}
	case OPC_MEMORY:
		disp := p.To.Offset

		if (disp > 0 && disp&0xffff != disp) ||
			(disp < 0 && -disp&0xffff != -disp) {
			c.ctxt.Diag("Memory instruction with immediate(0x%x) bigger than 0xffff at\n%v\n", disp, p)
		}

		o1 = OP_MEMORY(op.OPCode,
			getRegister(p.From.Reg), getRegister(p.To.Reg), int16(disp))

	case OPC_MISI_MEMORY:
		o1 = OP_MISI_MEMORY(op.OPCode,
			uint32(op.FnCode), getRegister(p.From.Reg), getRegister(p.To.Reg))

	case OPC_FUNC_MEMORY:
		disp := p.To.Offset

		if (disp >= 0x1000) || (disp <= -0x1000) {
			c.ctxt.Diag("Memory instruction with immediate(0x%x) bigger than 0x1000 at\n%v\n", disp, p)
		}

		o1 = OP_FUNC_MEMORY(op.OPCode,
			uint32(op.FnCode), getRegister(p.From.Reg), getRegister(p.To.Reg), int16(disp))

	case OPC_ARITHMETIC:
		o1 = OP_ARITHMETIC(op.OPCode,
			uint32(op.FnCode), getRegister(p.From.Reg), getReg2(p), getRegister(p.To.Reg))
	case OPC_ARITHMETIC_I:
		ib := p.RestArgs[0].Offset
		if ib > 255 {
			c.ctxt.Diag("Arithmetic instruction with immediate(0x%x) bigger than 0xff at\n%v\n", ib, p)
		}
		o1 = OP_ARITHMETIC_I(op.OPCode, uint32(op.FnCode), getRegister(p.From.Reg),
			int16(ib), getRegister(p.To.Reg))

	case OPC_COMPLEX_ARITHMETIC:
		o1 = OP_COMPLEX_ARITHMETIC(op.OPCode,
			uint32(op.FnCode), getRegister(p.From.Reg), getReg2(p), getFrom3Reg(p), getRegister(p.To.Reg))

	default:
		p.Ctxt.Diag("sw64 todo implement %v", p)
	}
	return []uint32{o1}
}

type OPClass uint8

const (
	OPC_NULL    = iota // 类别         |31  26|25 21|20 16|15       5|4   0|
	OPC_SYSCALL        //系统调用指令, |Opcode|         Function           |
	OPC_CONTROL        //转移控制指令, |Opcode| Ra  |       disp           |
	OPC_MEMORY         //存储器指令,   |Opcode| Ra  | Rb  |  disp          |

	OPC_MISI_MEMORY // 杂项指令     |Opcode| Ra  | Rb  |  Function      |

	OPC_FUNC_MEMORY //带功能域的存  |Opcode| Ra  | Rb  | Function |  disp   | // ZHAIAYH
	//储器指令格式  |31  26|25 21|20 16|15  12    |11      0|

	OPC_ARITHMETIC   //简单运算指令, |Opcode| Ra  | Rb  | Function | RC  |
	OPC_ARITHMETIC_I //简单运算指令, |Opcode| Ra  | Ib    |Function| RC  |
	//                                      |31  26|25 21|20   13|12     5|4   0|

	OPC_COMPLEX_ARITHMETIC   // 浮点复合运算指令格式 |Opcode|  Fa  |  Fb  |Function | Fc  |Fd  |
	OPC_COMPLEX_ARITHMETIC_I // 浮点复合运算指令格式 |Opcode|  Fa  |  Fb  |Function | Ib  |Fd  |
	//                      |31  26|25  21|20  16|15     10|9   5|4  0|
)

type OPTable struct {
	OPCode  uint8
	FnCode  uint16
	AS      obj.As
	Class   OPClass
	Comment string
}

func rbIsImmediate(p *obj.Prog) (bool, error) {
	if len(p.RestArgs) != 1 {
		return false, nil
	}
	a := p.RestArgs[0]
	switch a.Type {
	case obj.TYPE_CONST:
		return true, nil
	case obj.TYPE_REG:
		return false, nil
	}
	return false, fmt.Errorf("illegal RestArgs[0] %v", p.RestArgs[0])
}

func getRegister(v int16) int16 {
	if v == 0 {
		return REGZERO & 31
	}
	return v & 31
}
func getReg2(p *obj.Prog) int16 {
	reg := p.Reg
	if reg == 0 && p.RestArgs != nil {
		reg = p.RestArgs[0].Reg
	}
	return reg & 31
}

func getFrom3Reg(p *obj.Prog) int16 {
	reg := p.RestArgs[1].Reg
	if reg == 0 {
		panic(fmt.Sprintf("illegal From3 reg: %v", reg))
	}
	return reg & 31
}

var debug = false

func OP_SYSCALL(op uint8, fn uint32) uint32 {
	//            |31  26|25 21|20 16|15       5|4   0|
	//系统调用指令, |Opcode|         Function           |
	if debug {
		fmt.Printf("op:%x, fn:%x\n", op, fn)
	}
	return uint32(op)<<26 + fn
}

func OP_CONTROL(op uint8, ra int16, disp int32) uint32 {
	//            |31  26|25 21|20                   0|
	//转移控制指令, |Opcode| Ra  |       disp           |
	if debug {
		fmt.Printf("op:%x, ra:%x disp:%x\n", op, ra, uint32(disp)&(1<<21-1))
	}
	return uint32(op)<<26 + uint32(ra)<<21 + uint32(disp)&(1<<21-1)
}
func OP_MEMORY(op uint8, ra int16, rb int16, disp int16) uint32 {
	//            |31  26|25 21|20 16|15             0|
	//存储器指令    |Opcode| Ra  | Rb  |  disp          |
	if debug {
		fmt.Printf("op:%x, ra:%x rb:%x disp:%x\n", op, ra, rb, uint16(disp))
	}
	return uint32(op)<<26 + uint32(ra)<<21 + uint32(rb)<<16 + uint32(uint16(disp))
}

func OP_MISI_MEMORY(op uint8, fn uint32, ra int16, rb int16) uint32 {
	//              |31  26|25 21|20 16|15             0|
	//存储器指令    |Opcode| Ra  | Rb  |  FUNC          |
	if debug {
		fmt.Printf("op:%x, fn:%x, ra:%x rb:%x\n", op, uint16(fn), ra, rb)
	}
	return uint32(op)<<26 + uint32(ra)<<21 + uint32(rb)<<16 + uint32(uint16(fn))
}

func OP_FUNC_MEMORY(op uint8, fn uint32, ra int16, rb int16, disp int16) uint32 {
	//                        |31  26|25 21|20 16|15   12|11      0|
	//带功能域的存储器指令    |Opcode| Ra  | Rb  | FUNC  |  disp   |
	if debug {
		fmt.Printf("op:%x, fn:%x, ra:%x rb:%x disp:%x\n", op, fn, ra, rb, uint16(disp))
	}
	return uint32(op)<<26 + uint32(ra)<<21 + uint32(rb)<<16 + uint32(fn)<<12 + uint32(uint16(disp))
}

func OP_ARITHMETIC(op uint8, fn uint32, ra int16, rb int16, rc int16) uint32 {
	//            |31  26|25 21|20 16|15       5|4   0|
	//简单运算指令, |Opcode| Ra  | Rb  | Function | RC  |
	if debug {
		fmt.Printf("op:%x, ra:%x, rb:%x, fn:%x, rc:%x\n",
			op, ra, rb, fn, rc)
	}
	return uint32(op)<<26 + uint32(ra)<<21 + uint32(rb)<<16 + uint32(fn)<<5 + uint32(rc)
}
func OP_ARITHMETIC_I(op uint8, fn uint32, ra int16, ib int16, rc int16) uint32 {
	//            |31  26|25 21|20   13|12     5|4   0|
	//简单运算指令, |Opcode| Ra  | Ib    |Function| RC  |
	if debug {
		fmt.Printf("op:%x, ra:%x, ib:%x, fn:%x, rc:%x\n",
			op, ra, ib, fn, rc)
	}
	return uint32(op)<<26 + uint32(ra)<<21 + uint32(ib)<<13 + uint32(fn)<<5 + uint32(rc)
}

func OP_COMPLEX_ARITHMETIC(op uint8, fn uint32, ra int16, rb int16, rc int16, rd int16) uint32 {
	//                  |31  26|25 21|20 16|15      10|9   5|4  0|
	//浮点复合运算指令, |Opcode| Ra  | Rb  | Function | Rc  | Rd |
	if debug {
		fmt.Printf("op:%x, ra:%x, rb:%x, fn:%x, rc:%x, rd:%x \n",
			op, ra, rb, fn, rc, rd)
	}
	return uint32(op)<<26 + uint32(ra)<<21 + uint32(rb)<<16 + uint32(fn)<<10 + uint32(rc)<<5 + uint32(rd)
}

func addrel(cursym *obj.LSym, off int32, add int64, target *obj.LSym, rt objabi.RelocType) {
	r := obj.Addrel(cursym)
	r.Add = add
	r.Siz = 4
	r.Off = off
	r.Sym = target
	r.Type = rt
}

func op(a obj.As) uint8 {
	for _, i := range optabs {
		if i.AS == a {
			return i.OPCode
		}
	}
	panic("Unknown As")
}

func findOPTable(p *obj.Prog) (OPTable, error) {
	//snyh_TODO: avoid loop optabs dynamic
	ib, err := rbIsImmediate(p)
	if err != nil {
		return OPTable{}, err
	}
	as := p.As
	for _, o := range optabs {
		if as != o.AS {
			continue
		}
		switch o.Class {
		case OPC_ARITHMETIC_I:
			if ib == true {
				return o, nil
			}
		case OPC_ARITHMETIC:
			if ib == false {
				return o, nil
			}
		default:
			return o, nil
		}
	}
	return OPTable{}, fmt.Errorf("Can't find optable for %v", as)
}

var optabs = []OPTable{
	//杂项指令 snyh_TODO
	//锁指令 snyh_TODO
	//特权指令 snyh_TODO
	//原子操作指令 snyh_TODO
	//不可cache访存指令 snyh_TODO
	//浮点数指令 snyh_TODO
	//SMID指令 snyh_TODO

	//系统调用指令
	{0x00, 0x0, ASYS_CALL_B, OPC_SYSCALL, "带栏栅的调用特权程序"},
	{0x00, 0x1, ASYS_CALL, OPC_SYSCALL, "不带栏栅的调用特权程序"},

	{0x00, 0x0, obj.AUNDEF, OPC_SYSCALL, "undefined instruction"},

	{0x10, 0x0, AEXCB, OPC_MEMORY, "exception barrier"},

	//跳转指令
	{0x01, 0, ACALL, OPC_MEMORY, "跳转到子程序"},
	{0x02, 0, ARET, OPC_MEMORY, "子程序返回"},
	{0x03, 0, AJMP, OPC_MEMORY, "跳转"},
	//转移指令
	{0x04, 0, ABR, OPC_CONTROL, "无条件转移"},
	{0x05, 0, ABSR, OPC_CONTROL, "转到子程序"},

	{0x06, 0x0000, AMEMB, OPC_MISI_MEMORY, "存储器栏栅"},
	{0x06, 0x0001, AIMEMB, OPC_MISI_MEMORY, "指令栏栅"},
	{0x06, 0x1000, ARD_F, OPC_MISI_MEMORY, "读锁标志"},
	{0x06, 0x1020, AWR_F, OPC_MISI_MEMORY, "写锁标志"},
	{0x08, 0x0, ALLDW, OPC_FUNC_MEMORY, "锁装入字"},
	{0x08, 0x1, ALLDL, OPC_FUNC_MEMORY, "锁装入长字"},
	{0x08, 0x8, ALSTW, OPC_FUNC_MEMORY, "锁存储字"},
	{0x08, 0x9, ALSTL, OPC_FUNC_MEMORY, "锁存储长字"},

	{0x10, 0x00, AADDW, OPC_ARITHMETIC, "字加"},
	{0x10, 0x01, ASUBW, OPC_ARITHMETIC, "字减"},
	{0x10, 0x02, AS4ADDW, OPC_ARITHMETIC, "乘4的字加"},
	{0x10, 0x03, AS4SUBW, OPC_ARITHMETIC, "乘4的字减"},
	{0x10, 0x04, AS8ADDW, OPC_ARITHMETIC, "乘8的字加"},
	{0x10, 0x05, AS8SUBW, OPC_ARITHMETIC, "乘8的字减"},
	{0x10, 0x08, AADDL, OPC_ARITHMETIC, "长字加"},
	{0x10, 0x09, ASUBL, OPC_ARITHMETIC, "长字减"},
	{0x10, 0x0a, AS4ADDL, OPC_ARITHMETIC, "乘4的长字加"},
	{0x10, 0x0b, AS4SUBL, OPC_ARITHMETIC, "乘4的长字减"},
	{0x10, 0x0c, AS8ADDL, OPC_ARITHMETIC, "乘8的长字加"},
	{0x10, 0x0d, AS8SUBL, OPC_ARITHMETIC, "乘8的长字减"},
	{0x10, 0x10, AMULW, OPC_ARITHMETIC, "字乘"},
	{0x10, 0x18, AMULL, OPC_ARITHMETIC, "长字乘"},
	{0x10, 0x19, AUMULH, OPC_ARITHMETIC, "无符号数乘高64位"},
	{0x10, 0x28, ACMPEQ, OPC_ARITHMETIC, "有符号数等于比较"},
	{0x10, 0x29, ACMPLT, OPC_ARITHMETIC, "有符号数小于比较"},
	{0x10, 0x2a, ACMPLE, OPC_ARITHMETIC, "有符号数小于等于比较"},
	{0x10, 0x2b, ACMPULT, OPC_ARITHMETIC, "无符号数小于比较"},
	{0x10, 0x2c, ACMPULE, OPC_ARITHMETIC, "无符号数小于等于比较"},
	{0x10, 0x38, AAND, OPC_ARITHMETIC, "逻辑与"},
	{0x10, 0x39, ABIC, OPC_ARITHMETIC, "逻辑与非"},
	{0x10, 0x3a, ABIS, OPC_ARITHMETIC, "逻辑或"},
	{0x10, 0x3b, AORNOT, OPC_ARITHMETIC, "逻辑或非"},
	{0x10, 0x3c, AXOR, OPC_ARITHMETIC, "逻辑异或"},
	{0x10, 0x3d, AEQV, OPC_ARITHMETIC, "逻辑等效"},
	{0x10, 0x40, AINSLB, OPC_ARITHMETIC, "低段字节插入"},
	{0x10, 0x41, AINSLH, OPC_ARITHMETIC, "低段半字插入"},
	{0x10, 0x42, AINSLW, OPC_ARITHMETIC, "低段字插入"},
	{0x10, 0x43, AINSLL, OPC_ARITHMETIC, "低段长字插入"},
	{0x10, 0x44, AINSHB, OPC_ARITHMETIC, "高段字节插入"},
	{0x10, 0x45, AINSHH, OPC_ARITHMETIC, "高段半字插入"},
	{0x10, 0x46, AINSHW, OPC_ARITHMETIC, "高段字插入"},
	{0x10, 0x47, AINSHL, OPC_ARITHMETIC, "高段长字插入"},
	{0x10, 0x48, ASLL, OPC_ARITHMETIC, "逻辑左移"},
	{0x10, 0x49, ASRL, OPC_ARITHMETIC, "逻辑左移"},
	{0x10, 0x4a, ASRA, OPC_ARITHMETIC, "算术右移"},
	{0x10, 0x50, AEXTLB, OPC_ARITHMETIC, "低段字节提取"},
	{0x10, 0x51, AEXTLH, OPC_ARITHMETIC, "低段半字提取"},
	{0x10, 0x52, AEXTLW, OPC_ARITHMETIC, "低段字提取"},
	{0x10, 0x53, AEXTLL, OPC_ARITHMETIC, "低段长字提取"},
	{0x10, 0x54, AEXTHB, OPC_ARITHMETIC, "高段字节提取"},
	{0x10, 0x55, AEXTHH, OPC_ARITHMETIC, "高段半字提取"},
	{0x10, 0x56, AEXTHW, OPC_ARITHMETIC, "高段字提取"},
	{0x10, 0x57, AEXTHL, OPC_ARITHMETIC, "高段长字提取"},
	{0x10, 0x58, ACTPOP, OPC_ARITHMETIC, "数1"},
	{0x10, 0x59, ACTLZ, OPC_ARITHMETIC, "数头0"},
	{0x10, 0x5a, ACTTZ, OPC_ARITHMETIC, "数尾0"},
	{0x10, 0x60, AMASKLB, OPC_ARITHMETIC, "低段字节屏蔽"},
	{0x10, 0x61, AMASKLH, OPC_ARITHMETIC, "低段半字屏蔽"},
	{0x10, 0x62, AMASKLW, OPC_ARITHMETIC, "低段字屏蔽"},
	{0x10, 0x63, AMASKLL, OPC_ARITHMETIC, "低段长字屏蔽"},
	{0x10, 0x64, AMASKHB, OPC_ARITHMETIC, "高段字节屏蔽"},
	{0x10, 0x65, AMASKHH, OPC_ARITHMETIC, "高段半字屏蔽"},
	{0x10, 0x66, AMASKHW, OPC_ARITHMETIC, "高段字屏蔽"},
	{0x10, 0x67, AMASKHL, OPC_ARITHMETIC, "高段长字屏蔽"},
	{0x10, 0x68, AZAP, OPC_ARITHMETIC, "字节清除"},
	{0x10, 0x69, AZAPNOT, OPC_ARITHMETIC, "字节反码清除"},
	{0x10, 0x6a, ASEXTB, OPC_ARITHMETIC, "字节符号扩展"},
	{0x10, 0x6b, ASEXTH, OPC_ARITHMETIC, "半字符号扩展"},
	{0x10, 0x6c, ACMPGEB, OPC_ARITHMETIC, "字节无符号比较"},

	{0x10, 0x70, AFIMOVS, OPC_ARITHMETIC, "S浮点传送到整数"},
	{0x10, 0x78, AFIMOVD, OPC_ARITHMETIC, "D浮点传送到整数"},

	//寄存器格式的整数条件选择指令
	{0x11, 0x0, ASELEQ, OPC_ARITHMETIC, "整数等于0选"},
	{0x11, 0x1, ASELGE, OPC_ARITHMETIC, "整数大于等于0选"},
	{0x11, 0x2, ASELGT, OPC_ARITHMETIC, "整数大于0择"},
	{0x11, 0x3, ASELLE, OPC_ARITHMETIC, "整数小于等于0选"},
	{0x11, 0x4, ASELLT, OPC_ARITHMETIC, "整数小于0选"},
	{0x11, 0x5, ASELNE, OPC_ARITHMETIC, "整数不等于0选择"},
	{0x11, 0x6, ASELLBC, OPC_ARITHMETIC, "整数最低位为0选"},
	{0x11, 0x7, ASELLBS, OPC_ARITHMETIC, "整数最低位为1选"},

	//立即数格式的整数运算指令
	{0x12, 0x00, AADDW, OPC_ARITHMETIC_I, "字加"},
	{0x12, 0x01, ASUBW, OPC_ARITHMETIC_I, "字减"},
	{0x12, 0x02, AS4ADDW, OPC_ARITHMETIC_I, "乘4的字加"},
	{0x12, 0x03, AS4SUBW, OPC_ARITHMETIC_I, "乘4的字减"},
	{0x12, 0x04, AS8ADDW, OPC_ARITHMETIC_I, "乘8的字加"},
	{0x12, 0x05, AS8SUBW, OPC_ARITHMETIC_I, "乘8的字减"},
	{0x12, 0x08, AADDL, OPC_ARITHMETIC_I, "长字加"},
	{0x12, 0x09, ASUBL, OPC_ARITHMETIC_I, "长字减"},
	{0x12, 0x0a, AS4ADDL, OPC_ARITHMETIC_I, "乘4的长字加"},
	{0x12, 0x0b, AS4SUBL, OPC_ARITHMETIC_I, "乘4的长字减"},
	{0x12, 0x0c, AS8ADDL, OPC_ARITHMETIC_I, "乘8的长字加"},
	{0x12, 0x0d, AS8SUBL, OPC_ARITHMETIC_I, "乘8的长字减"},
	{0x12, 0x10, AMULW, OPC_ARITHMETIC_I, "字乘"},
	{0x12, 0x18, AMULL, OPC_ARITHMETIC_I, "长字乘"},
	{0x12, 0x19, AUMULH, OPC_ARITHMETIC_I, "无符号数乘高64位"},
	{0x12, 0x28, ACMPEQ, OPC_ARITHMETIC_I, "有符号数等于比较"},
	{0x12, 0x29, ACMPLT, OPC_ARITHMETIC_I, "有符号数小于比较"},
	{0x12, 0x2a, ACMPLE, OPC_ARITHMETIC_I, "有符号数小于等于比较"},
	{0x12, 0x2b, ACMPULT, OPC_ARITHMETIC_I, "无符号数小于比较"},
	{0x12, 0x2c, ACMPULE, OPC_ARITHMETIC_I, "无符号数小于等于比较"},
	{0x12, 0x38, AAND, OPC_ARITHMETIC_I, "逻辑与"},
	{0x12, 0x39, ABIC, OPC_ARITHMETIC_I, "逻辑与非"},
	{0x12, 0x3a, ABIS, OPC_ARITHMETIC_I, "逻辑或"},
	{0x12, 0x3b, AORNOT, OPC_ARITHMETIC_I, "逻辑或非"},
	{0x12, 0x3c, AXOR, OPC_ARITHMETIC_I, "逻辑异或"},
	{0x12, 0x3d, AEQV, OPC_ARITHMETIC_I, "逻辑等效"},
	{0x12, 0x40, AINSLB, OPC_ARITHMETIC_I, "低段字节插入"},
	{0x12, 0x41, AINSLH, OPC_ARITHMETIC_I, "低段半字插入"},
	{0x12, 0x42, AINSLW, OPC_ARITHMETIC_I, "低段字插入"},
	{0x12, 0x43, AINSLL, OPC_ARITHMETIC_I, "低段长字插入"},
	{0x12, 0x44, AINSHB, OPC_ARITHMETIC_I, "高段字节插入"},
	{0x12, 0x45, AINSHH, OPC_ARITHMETIC_I, "高段半字插入"},
	{0x12, 0x46, AINSHW, OPC_ARITHMETIC_I, "高段字插入"},
	{0x12, 0x47, AINSHL, OPC_ARITHMETIC_I, "高段长字插入"},
	{0x12, 0x48, ASLL, OPC_ARITHMETIC_I, "逻辑左移"},
	{0x12, 0x49, ASRL, OPC_ARITHMETIC_I, "逻辑左移"},
	{0x12, 0x4a, ASRA, OPC_ARITHMETIC_I, "算术右移"},
	{0x12, 0x50, AEXTLB, OPC_ARITHMETIC_I, "低段字节提取"},
	{0x12, 0x51, AEXTLH, OPC_ARITHMETIC_I, "低段半字提取"},
	{0x12, 0x52, AEXTLW, OPC_ARITHMETIC_I, "低段字提取"},
	{0x12, 0x53, AEXTLL, OPC_ARITHMETIC_I, "低段长字提取"},
	{0x12, 0x54, AEXTHB, OPC_ARITHMETIC_I, "高段字节提取"},
	{0x12, 0x55, AEXTHH, OPC_ARITHMETIC_I, "高段半字提取"},
	{0x12, 0x56, AEXTHW, OPC_ARITHMETIC_I, "高段字提取"},
	{0x12, 0x57, AEXTHL, OPC_ARITHMETIC_I, "高段长字提取"},
	{0x12, 0x60, AMASKLB, OPC_ARITHMETIC_I, "低段字节屏蔽"},
	{0x12, 0x61, AMASKLH, OPC_ARITHMETIC_I, "低段半字屏蔽"},
	{0x12, 0x62, AMASKLW, OPC_ARITHMETIC_I, "低段字屏蔽"},
	{0x12, 0x63, AMASKLL, OPC_ARITHMETIC_I, "低段长字屏蔽"},
	{0x12, 0x64, AMASKHB, OPC_ARITHMETIC_I, "高段字节屏蔽"},
	{0x12, 0x65, AMASKHH, OPC_ARITHMETIC_I, "高段半字屏蔽"},
	{0x12, 0x66, AMASKHW, OPC_ARITHMETIC_I, "高段字屏蔽"},
	{0x12, 0x67, AMASKHL, OPC_ARITHMETIC_I, "高段长字屏蔽"},
	{0x12, 0x68, AZAP, OPC_ARITHMETIC_I, "字节清除"},
	{0x12, 0x69, AZAPNOT, OPC_ARITHMETIC_I, "字节反码清除"},
	{0x12, 0x6a, ASEXTB, OPC_ARITHMETIC_I, "字节符号扩展"},
	{0x12, 0x6b, ASEXTH, OPC_ARITHMETIC_I, "半字符号扩展"},
	{0x12, 0x6c, ACMPGEB, OPC_ARITHMETIC_I, "字节无符号比较"},
	//立即数格式的整数条件选择指令
	{0x13, 0x0, ASELEQ, OPC_ARITHMETIC_I, "整数等于0选"},
	{0x13, 0x1, ASELGE, OPC_ARITHMETIC_I, "整数大于等于0选"},
	{0x13, 0x2, ASELGT, OPC_ARITHMETIC_I, "整数大于0择"},
	{0x13, 0x3, ASELLE, OPC_ARITHMETIC_I, "整数小于等于0选"},
	{0x13, 0x4, ASELLT, OPC_ARITHMETIC_I, "整数小于0选"},
	{0x13, 0x5, ASELNE, OPC_ARITHMETIC_I, "整数不等于0选择"},
	{0x13, 0x6, ASELLBC, OPC_ARITHMETIC_I, "整数最低位为0选"},
	{0x13, 0x7, ASELLBS, OPC_ARITHMETIC_I, "整数最低位为1选"},

	//存储装入指令
	{0x20, 0, ALDBU, OPC_MEMORY, "装入字节并作零扩展"},
	{0x21, 0, ALDHU, OPC_MEMORY, "装入半字并作零扩展"},
	{0x22, 0, ALDW, OPC_MEMORY, "装入字"},
	{0x23, 0, ALDL, OPC_MEMORY, "装入长字"},
	{0x24, 0, ALDL_U, OPC_MEMORY, "装入不对界的长字"},

	//特权指令
	{0x25, 0, APRI_LD, OPC_MEMORY, "特权装入指令"},

	//存储装入指令
	{0x26, 0, AFLDS, OPC_MEMORY, "装入S浮点或字"},
	{0x27, 0, AFLDD, OPC_MEMORY, "装入D浮点"},
	{0x28, 0, ASTB, OPC_MEMORY, "存储字节"},
	{0x29, 0, ASTH, OPC_MEMORY, "存储半字"},
	{0x2A, 0, ASTW, OPC_MEMORY, "存储字"},
	{0x2B, 0, ASTL, OPC_MEMORY, "存储长字"},
	{0x2C, 0, ASTL_U, OPC_MEMORY, "U存储不对界长字"},
	{0x2D, 0, APRI_ST, OPC_MEMORY, "特权存储指令"},
	{0x2E, 0, AFSTS, OPC_MEMORY, "存储S浮点或字"},
	{0x2F, 0, AFSTD, OPC_MEMORY, "存储D浮点"},

	//转移指令
	{0x30, 0, ABEQ, OPC_CONTROL, "等于0转移"},
	{0x31, 0, ABNE, OPC_CONTROL, "不等于0转移"},
	{0x32, 0, ABLT, OPC_CONTROL, "小于0转移"},
	{0x33, 0, ABLE, OPC_CONTROL, "小于等于0转移"},
	{0x34, 0, ABGT, OPC_CONTROL, "大于0转移"},
	{0x35, 0, ABGE, OPC_CONTROL, "大于等于0转移"},
	{0x36, 0, ABLBC, OPC_CONTROL, "最低位为0转移"},
	{0x37, 0, ABLBS, OPC_CONTROL, "最低位为1转移"},
	{0x38, 0, AFBEQ, OPC_CONTROL, "浮点等于0转移"},
	{0x39, 0, AFBNE, OPC_CONTROL, "浮点不等于0转移"},
	{0x3A, 0, AFBLT, OPC_CONTROL, "浮点小于0转移"},
	{0x3B, 0, AFBLE, OPC_CONTROL, "浮点小于等于0转移"},
	{0x3C, 0, AFBGT, OPC_CONTROL, "浮点大于0转移"},
	{0x3D, 0, AFBGE, OPC_CONTROL, "浮点大于等于0转移"},

	//装入立即数指令
	{0x3e, 0, ALDI, OPC_MEMORY, "装入立即数"},
	{0x3f, 0, ALDIH, OPC_MEMORY, "装入立即数到高位"},

	{0x18, 0x00, AFADDS, OPC_ARITHMETIC, "S浮点加"},
	{0x18, 0x01, AFADDD, OPC_ARITHMETIC, "D浮点加"},
	{0x18, 0x02, AFSUBS, OPC_ARITHMETIC, "S浮点减"},
	{0x18, 0x03, AFSUBD, OPC_ARITHMETIC, "D浮点减"},
	{0x18, 0x04, AFMULS, OPC_ARITHMETIC, "S浮点乘"},
	{0x18, 0x05, AFMULD, OPC_ARITHMETIC, "D浮点乘"},
	{0x18, 0x06, AFDIVS, OPC_ARITHMETIC, "S浮点除"},
	{0x18, 0x07, AFDIVD, OPC_ARITHMETIC, "D浮点除"},
	{0x18, 0x08, AFSQRTS, OPC_ARITHMETIC, "S浮点平方根"},
	{0x18, 0x09, AFSQRTD, OPC_ARITHMETIC, "D浮点平方根"},
	{0x18, 0x10, AFCMPEQ, OPC_ARITHMETIC, "浮点等于比较"},
	{0x18, 0x11, AFCMPLE, OPC_ARITHMETIC, "浮点小于等于比较"},
	{0x18, 0x12, AFCMPLT, OPC_ARITHMETIC, "浮点小于比较"},
	{0x18, 0x13, AFCMPUN, OPC_ARITHMETIC, "浮点不可排序比较"},
	{0x18, 0x20, AFCVTSD, OPC_ARITHMETIC, "S浮点转为D浮点"},
	{0x18, 0x21, AFCVTDS, OPC_ARITHMETIC, "D浮点转为S浮点"},
	{0x18, 0x22, AFCVTDL_G, OPC_ARITHMETIC, "FCVTDL就近舍入"},
	{0x18, 0x23, AFCVTDL_P, OPC_ARITHMETIC, "FCVTDL向正无穷舍入"},
	{0x18, 0x24, AFCVTDL_Z, OPC_ARITHMETIC, "FCVTDL向“0”舍入"},
	{0x18, 0x25, AFCVTDL_N, OPC_ARITHMETIC, "FCVTDL向负无穷舍入"},
	{0x18, 0x27, AFCVTDL, OPC_ARITHMETIC, "D浮点转为长字"},
	{0x18, 0x28, AFCVTWL, OPC_ARITHMETIC, "字转为长字"},
	{0x18, 0x29, AFCVTLW, OPC_ARITHMETIC, "长字转为字"},
	{0x18, 0x2D, AFCVTLS, OPC_ARITHMETIC, "长字转S浮点"},
	{0x18, 0x2F, AFCVTLD, OPC_ARITHMETIC, "长字转D浮点"},
	{0x18, 0x30, AFCPYS, OPC_ARITHMETIC, "拷贝符号"},
	{0x18, 0x31, AFCPYSE, OPC_ARITHMETIC, "拷贝符号和指数"},
	{0x18, 0x32, AFCPYSN, OPC_ARITHMETIC, "拷贝符号的反码"},
	{0x18, 0x40, AIFMOVS, OPC_ARITHMETIC, "整数传送到S浮点"},
	{0x18, 0x41, AIFMOVD, OPC_ARITHMETIC, "整数传送到D浮点"},
	{0x18, 0x50, ARFPCR, OPC_ARITHMETIC, "读浮点控制寄存器FPCR"},
	{0x18, 0x51, AWFPCR, OPC_ARITHMETIC, "写浮点控制寄存器FPCR"},
	{0x18, 0x54, ASETFPEC0, OPC_ARITHMETIC, "设置FPCR异常模式位为00"},
	{0x18, 0x55, ASETFPEC1, OPC_ARITHMETIC, "设置FPCR异常模式位为01"},
	{0x18, 0x56, ASETFPEC2, OPC_ARITHMETIC, "设置FPCR异常模式位为10"},
	{0x18, 0x57, ASETFPEC3, OPC_ARITHMETIC, "设置FPCR异常模式位为11"},

	{0x19, 0x00, AFMAS, OPC_COMPLEX_ARITHMETIC, "S浮点乘加"},
	{0x19, 0x01, AFMAD, OPC_COMPLEX_ARITHMETIC, "D浮点乘加"},
	{0x19, 0x02, AFMSS, OPC_COMPLEX_ARITHMETIC, "S浮点乘减"},
	{0x19, 0x03, AFMSD, OPC_COMPLEX_ARITHMETIC, "D浮点乘减"},
	{0x19, 0x04, AFNMAS, OPC_COMPLEX_ARITHMETIC, "S浮点负乘加"},
	{0x19, 0x05, AFNMAD, OPC_COMPLEX_ARITHMETIC, "D浮点负乘加"},
	{0x19, 0x06, AFNMSS, OPC_COMPLEX_ARITHMETIC, "S浮点负乘减"},
	{0x19, 0x07, AFNMSD, OPC_COMPLEX_ARITHMETIC, "D浮点负乘减"},
}
